from bitcoding import B_dict
from hornsplit import hsplit as split
from inverse import inv
from solve import solve
#import random
from vbweights import w
from bitcoding import B_dict
from array import array
from counters import counter
from saturate import saturate
from triangulation import MCS, FIC
import time

allowance = 120 # Timeout; set to None for no timeout (note that by the time bnb is called, some time will have been wasted on greedy saturation, but that is negligible in general)

DALL = B_dict['DALL']

cardmult = max(w)

repairs = 0
variables = set([])
occurences = dict([])
conflicts = dict([])
lfConMatrix = None

def select(ConMatrix, models):
   try:
      selection = [(i,j) for i,j in conflicts if (i,j) in variables]
      if not selection:
         dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in variables])
      else:
         dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/float(conflicts[(i,j)]),(i,j)) for i,j in selection])
   except ValueError:
      return False
   return dummy, (i,j)

def bnb(ConMatrix, ConMatrixT, neighbors, chordal, pc_only, bound, cutoff, i, j, models, start, check = False):

   global repairs
   global allowance
   global lfConMatrix
   global conflicts
   global occurences
   global variables

   if start == -1:
      start = time.process_time() # or perf_counter() 
   
   if allowance and time.process_time() - start > allowance:
      yield bound, lfConMatrix, True
      return
   
   if bound <= cutoff:
      return

   if repairs >= bound: 
      return

   solution = ConMatrixT
   if check:
      solution, fail = solve(tuple([c[:] for c in ConMatrixT]), neighbors, chordal, pc_only)
      counter['oracle'] += 1
   
   if not solution: 
      conflicts[(i,j)] = counter['oracle']*occurences[(i,j)]
      conflicts[fail] =  counter['oracle']*occurences[fail]
      occurences[(i,j)]+=1
      occurences[fail]+=1
      return
      
   myNextVar = select(ConMatrix, models)
   if myNextVar != False:
      dummy, (i,j) = myNextVar
   else:
      bound = repairs
      yield bound, ConMatrixT, False
      return
         
   values = [DALL, ConMatrix[i][j]]
   for val in values:
      ConMatrixT[i][j] = val
      ConMatrixT[j][i] = inv[val-1]
      if val != ConMatrix[i][j]: repairs +=1
      variables.discard((i,j))
      for bound, output, timeout in bnb(ConMatrix, ConMatrixT, neighbors, chordal, pc_only, bound, cutoff, i, j, models, start, True if val != DALL else False):
         if output != None:
            result = tuple([c[:] for c in output])
            lfConMatrix = result
            yield bound, result, timeout
      if val != ConMatrix[i][j]: repairs -= 1 
      variables.add((i,j))
      ConMatrixT[i][j] = DALL
      ConMatrixT[j][i] = DALL
         
   return

def findSaturation(ConMatrix, ConMatrixT, neighbors, chordal, pc_only):

   Vars = len(ConMatrix)
   variables.clear()
   conflicts.clear()
   occurences.clear()
   global repairs
   global lfConMatrix
   repairs = 0
   
   edges = set([])
   for i in range(Vars):
      for j in range(i+1,Vars):
         occurences[(i,j)] = 1
         if ConMatrix[i][j] != DALL:
             edges.add((i,j))  
             if chordal: 
                neighbors[i].add(j)
                neighbors[j].add(i)
   
   if chordal:
      a, a_ = MCS(ConMatrix, neighbors)
      fill = FIC(ConMatrix, neighbors, a, a_)

      for i,j in fill:
          neighbors[i].add(j)
          neighbors[j].add(i)                   

   portfolio = []
   fixes = None
   cutoff = 1 if not solve(tuple([c[:] for c in ConMatrix]), neighbors, chordal, pc_only)[0] else 0 # Number of repairs with which you would be happy; if the QCN is UNSAT, 1 is the lowest bound
   if not cutoff: 
      counter['oracle'] += 1
      return 0, ConMatrix, False
   for model in range(6+1,1-1,-1):
      fixes, saturation, flag = saturate(ConMatrix, tuple([c[:] for c in ConMatrixT]), set(edges), neighbors, chordal, pc_only, model, fixes)
      if not flag:
         portfolio.append((fixes, saturation, model))
      if fixes <= cutoff: break
      
   fixes, saturation, model = min(portfolio)    
   
   while edges:
      i,j = edges.pop()
      variables.add((i,j))

   k = -1 # This variable controls the following loop; setting it to -1 allows one to find lowest bound, obviously the loop will run at least once, so k should be >= 1 
   fcounter = 1
   bound, result, timeout = fixes, saturation, False
   lfConMatrix = saturation
   
   for bound, result, timeout in bnb(ConMatrix, tuple([c[:] for c in ConMatrixT]), neighbors, chordal, pc_only, fixes, cutoff, -1, -1, None, -1):
      if fcounter <= k: break
      fcounter += 1
      if timeout:
         break

   return bound, result, timeout
