from inverse import inv
from glob import globs
from vbweights import w
from counters import counter
from assignment import mrv_variables, initialize_mrv_vars, add_var, remove_var
from collections import deque
from learned_weights import learnedWeights

if globs['split'] == 0: # splitting based on set of base relations
   from basesplit import bsplit as split
elif globs['split'] == 1: # splitting based on horn set
   from hornsplit import hsplit as split

if globs['pcheuristic'] == 0: # path consistency
   from pc import PC as pc
elif globs['pcheuristic'] == 1: # path consistency with weighted queue
   from pcw import PCw as pc

if globs['heuristic'] == 0: # none
   def myvar(i_j, neighbors, ConMatrix):
      i, j = i_j
      for ci,cj in mrv_variables:
         dummy, (i,j) = 0, (ci,cj)
         return dummy, (i,j)
      return False
elif globs['heuristic'] == 2: # global
   def myvar(i_j, neighbors, ConMatrix):
      i, j = i_j
      try:
         selection = [(i,j) for i,j in learnedWeights if (i,j) in mrv_variables]
         if not selection:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in mrv_variables])
         else:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/float(learnedWeights[(i,j)]),(i,j)) for i,j in selection])
      except:
         return False
      return dummy, (i,j)

cardmult = max(w)

# iterative consistency
def iconsistency(ConMatrix, neighbors, chordal, m = -1, n = -1, repairs = None, trail = None):
   
   fail = ()
   
   stack = [] # initialize stack to simulate backtracking

   counter['nodes'] += 1

   lastAssigned = None

   initialize_mrv_vars(ConMatrix)
   learnedWeights.clear()

   # as long as the consistency problem is not decided, process it
   while 1:

      counter['decisions'] += 1

      if not lastAssigned:
         try:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in mrv_variables])
            lastAssigned = (i,j)
         except:
            return ConMatrix, fail
      else:
         myNextVar = myvar(lastAssigned, neighbors, ConMatrix)
         if myNextVar:
            dummy, (i,j) = myNextVar
            lastAssigned = (i,j)
         else:
            return ConMatrix, fail
 
      # splitting
      values = split[ConMatrix[i][j]-1][1][:]

      #least constraining value heuristic
      valuesw = [(-w[a-1],a) for a in values]
      valuesw.sort(reverse=True)
      valuesw = [a[1] for a in valuesw]

      #valuesw.reverse() # reverse list of values to get popped in the correct order
     
      # as long as a consistent variable-value pair in not found, search for it
      while 1:

         counter['nodes'] += 1 # increment visited nodes counter
 
         # check if current variable has any variables left, if not backtrack to a previous variable assignment
         if not valuesw:
            # check if any previous variable assignments are left in the stack
            while stack:
               repairs, trail, (i, j), valuesw, dummy = stack.pop()
               for g, h in repairs.keys():
                  ConMatrix[g][h] = repairs[(g, h)]
                  ConMatrix[h][g] = inv[repairs[(g, h)]-1]
               add_var((i,j))
               while trail:
                  c,(a,b) = trail.pop()
                  remove_var((a,b)) if c else add_var((a,b))
               if valuesw: 
                  break
            else:
              return None, fail         
               
         value = valuesw.pop() # grab first value from variable
         
         repairs = {}
         trail = deque([])

         repairs[(i,j)] = ConMatrix[i][j]

         # assignment takes place
         ConMatrix[i][j] = value
         ConMatrix[j][i] = inv[value-1]

         remove_var((i,j))
            
         # path consistency 
         pcresult, fail = pc(ConMatrix, neighbors, chordal, i, j, repairs, trail)
         if pcresult:
            break

         # revert contraint mantrix and mrv variables to previous state
         for g, h in repairs.keys():
            ConMatrix[g][h] = repairs[(g, h)] 
            ConMatrix[h][g] = inv[repairs[(g, h)]-1]
         add_var((i,j))
         while trail:
            c,(a,b) = trail.pop()
            remove_var((a,b)) if c else add_var((a,b))
  
      stack.append((repairs, trail, (i, j), valuesw[:], dummy)) # save current state (function call) in a stack
     
   raise RuntimeError("Can't happen")
